from itertools import chain
import os
import socket
import sys

from ledis._compat import (b, xrange, imap, byte_to_chr, unicode, bytes, long,
                           BytesIO, nativestr, basestring, iteritems,
                           LifoQueue, Empty, Full, urlparse, parse_qs)
from ledis.exceptions import (
    LedisError,
    ConnectionError,
    BusyLoadingError,
    ResponseError,
    InvalidResponse,
    ExecAbortError,
    )


SYM_STAR = b('*')
SYM_DOLLAR = b('$')
SYM_CRLF = b('\r\n')
SYM_LF = b('\n')


class PythonParser(object):
    "Plain Python parsing class"
    MAX_READ_LENGTH = 1000000
    encoding = None

    EXCEPTION_CLASSES = {
        'ERR': ResponseError,
        'EXECABORT': ExecAbortError,
        'LOADING': BusyLoadingError,
    }

    def __init__(self):
        self._fp = None

    def __del__(self):
        try:
            self.on_disconnect()
        except Exception:
            pass

    def on_connect(self, connection):
        "Called when the socket connects"
        self._fp = connection._sock.makefile('rb')
        if connection.decode_responses:
            self.encoding = connection.encoding

    def on_disconnect(self):
        "Called when the socket disconnects"
        if self._fp is not None:
            self._fp.close()
            self._fp = None

    def read(self, length=None):
        """
        Read a line from the socket if no length is specified,
        otherwise read ``length`` bytes. Always strip away the newlines.
        """
        try:
            if length is not None:
                bytes_left = length + 2  # read the line ending
                if length > self.MAX_READ_LENGTH:
                    # apparently reading more than 1MB or so from a windows
                    # socket can cause MemoryErrors. See:
                    # https://github.com/andymccurdy/redis-py/issues/205
                    # read smaller chunks at a time to work around this
                    try:
                        buf = BytesIO()
                        while bytes_left > 0:
                            read_len = min(bytes_left, self.MAX_READ_LENGTH)
                            buf.write(self._fp.read(read_len))
                            bytes_left -= read_len
                        buf.seek(0)
                        return buf.read(length)
                    finally:
                        buf.close()
                return self._fp.read(bytes_left)[:-2]

            # no length, read a full line
            return self._fp.readline()[:-2]
        except (socket.error, socket.timeout):
            e = sys.exc_info()[1]
            raise ConnectionError("Error while reading from socket: %s" %
                                  (e.args,))

    def parse_error(self, response):
        "Parse an error response"
        error_code = response.split(' ')[0]
        if error_code in self.EXCEPTION_CLASSES:
            response = response[len(error_code) + 1:]
            return self.EXCEPTION_CLASSES[error_code](response)
        return ResponseError(response)

    def read_response(self):
        response = self.read()
        if not response:
            raise ConnectionError("Socket closed on remote end")

        byte, response = byte_to_chr(response[0]), response[1:]

        if byte not in ('-', '+', ':', '$', '*'):
            raise InvalidResponse("Protocol Error")

        # server returned an error
        if byte == '-':
            response = nativestr(response)
            error = self.parse_error(response)
            # if the error is a ConnectionError, raise immediately so the user
            # is notified
            if isinstance(error, ConnectionError):
                raise error
            # otherwise, we're dealing with a ResponseError that might belong
            # inside a pipeline response. the connection's read_response()
            # and/or the pipeline's execute() will raise this error if
            # necessary, so just return the exception instance here.
            return error
        # single value
        elif byte == '+':
            pass
        # int value
        elif byte == ':':
            response = long(response)
        # bulk response
        elif byte == '$':
            length = int(response)
            if length == -1:
                return None
            response = self.read(length)
        # multi-bulk response
        elif byte == '*':
            length = int(response)
            if length == -1:
                return None
            response = [self.read_response() for i in xrange(length)]
        if isinstance(response, bytes) and self.encoding:
            response = response.decode(self.encoding)
        return response


DefaultParser = PythonParser


class Connection(object):
    "Manages TCP communication to and from a Ledis server"
    def __init__(self, host='localhost', port=6380, db=0, password=None,
                 socket_timeout=None, encoding='utf-8',
                 encoding_errors='strict', decode_responses=False,
                 parser_class=DefaultParser):
        self.pid = os.getpid()
        self.host = host
        self.port = port
        self.db = db
        self.password = password
        self.socket_timeout = socket_timeout
        self.encoding = encoding
        self.encoding_errors = encoding_errors
        self.decode_responses = decode_responses
        self._sock = None
        self._parser = parser_class()

    def __del__(self):
        try:
            self.disconnect()
        except Exception:
            pass

    def connect(self):
        "Connects to the Ledis server if not already connected"
        if self._sock:
            return
        try:
            sock = self._connect()
        except socket.error:
            e = sys.exc_info()[1]
            raise ConnectionError(self._error_message(e))

        self._sock = sock
        self.on_connect()

    def _connect(self):
        "Create a TCP socket connection"
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(self.socket_timeout)
        sock.connect((self.host, self.port))
        return sock

    def _error_message(self, exception):
        # args for socket.error can either be (errno, "message")
        # or just "message"
        if len(exception.args) == 1:
            return "Error connecting to %s:%s. %s." % \
                (self.host, self.port, exception.args[0])
        else:
            return "Error %s connecting %s:%s. %s." % \
                (exception.args[0], self.host, self.port, exception.args[1])

    def on_connect(self):
        "Initialize the connection, authenticate and select a database"
        self._parser.on_connect(self)

        # if a password is specified, authenticate
        if self.password:
            self.send_command('AUTH', self.password)
            if nativestr(self.read_response()) != 'OK':
                raise AuthenticationError('Invalid Password')

        # if a database is specified, switch to it
        if self.db:
            self.send_command('SELECT', self.db)
            if nativestr(self.read_response()) != 'OK':
                raise ConnectionError('Invalid Database')

    def disconnect(self):
        "Disconnects from the Ledis server"
        self._parser.on_disconnect()
        if self._sock is None:
            return
        try:
            self._sock.close()
        except socket.error:
            pass
        self._sock = None

    def send_packed_command(self, command):
        "Send an already packed command to the Ledis server"
        if not self._sock:
            self.connect()
        try:
            self._sock.sendall(command)
        except socket.error:
            e = sys.exc_info()[1]
            self.disconnect()
            if len(e.args) == 1:
                _errno, errmsg = 'UNKNOWN', e.args[0]
            else:
                _errno, errmsg = e.args
            raise ConnectionError("Error %s while writing to socket. %s." %
                                  (_errno, errmsg))
        except Exception:
            self.disconnect()
            raise

    def send_command(self, *args):
        "Pack and send a command to the Ledis server"
        self.send_packed_command(self.pack_command(*args))

    def read_response(self):
        "Read the response from a previously sent command"
        try:
            response = self._parser.read_response()
        except Exception:
            self.disconnect()
            raise
        if isinstance(response, ResponseError):
            raise response
        return response

    def encode(self, value):
        "Return a bytestring representation of the value"
        if isinstance(value, bytes):
            return value
        if isinstance(value, float):
            value = repr(value)
        if not isinstance(value, basestring):
            value = str(value)
        if isinstance(value, unicode):
            value = value.encode(self.encoding, self.encoding_errors)
        return value

    def pack_command(self, *args):
        "Pack a series of arguments into a value Ledis command"
        output = SYM_STAR + b(str(len(args))) + SYM_CRLF
        for enc_value in imap(self.encode, args):
            output += SYM_DOLLAR
            output += b(str(len(enc_value)))
            output += SYM_CRLF
            output += enc_value
            output += SYM_CRLF
        return output


class UnixDomainSocketConnection(Connection):
    def __init__(self, path='', db=0, password=None,
                 socket_timeout=None, encoding='utf-8',
                 encoding_errors='strict', decode_responses=False,
                 parser_class=DefaultParser):
        self.pid = os.getpid()
        self.path = path
        self.db = db
        self.password = password
        self.socket_timeout = socket_timeout
        self.encoding = encoding
        self.encoding_errors = encoding_errors
        self.decode_responses = decode_responses
        self._sock = None
        self._parser = parser_class()

    def _connect(self):
        "Create a Unix domain socket connection"
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        sock.settimeout(self.socket_timeout)
        sock.connect(self.path)
        return sock

    def _error_message(self, exception):
        # args for socket.error can either be (errno, "message")
        # or just "message"
        if len(exception.args) == 1:
            return "Error connecting to unix socket: %s. %s." % \
                (self.path, exception.args[0])
        else:
            return "Error %s connecting to unix socket: %s. %s." % \
                (exception.args[0], self.path, exception.args[1])


# TODO: add ability to block waiting on a connection to be released
class ConnectionPool(object):
    "Generic connection pool"
    @classmethod
    def from_url(cls, url, db=None, **kwargs):
        """
        Return a connection pool configured from the given URL.

        For example::

            redis://[:password]@localhost:6379/0
            rediss://[:password]@localhost:6379/0
            unix://[:password]@/path/to/socket.sock?db=0

        Three URL schemes are supported:
            redis:// creates a normal TCP socket connection
            rediss:// creates a SSL wrapped TCP socket connection
            unix:// creates a Unix Domain Socket connection

        There are several ways to specify a database number. The parse function
        will return the first specified option:
            1. A ``db`` querystring option, e.g. redis://localhost?db=0
            2. If using the redis:// scheme, the path argument of the url, e.g.
               redis://localhost/0
            3. The ``db`` argument to this function.

        If none of these options are specified, db=0 is used.

        Any additional querystring arguments and keyword arguments will be
        passed along to the ConnectionPool class's initializer. In the case
        of conflicting arguments, querystring arguments always win.
        """
        url_string = url
        url = urlparse(url)
        qs = ''

        # in python2.6, custom URL schemes don't recognize querystring values
        # they're left as part of the url.path.
        if '?' in url.path and not url.query:
            # chop the querystring including the ? off the end of the url
            # and reparse it.
            qs = url.path.split('?', 1)[1]
            url = urlparse(url_string[:-(len(qs) + 1)])
        else:
            qs = url.query

        url_options = {}

        for name, value in iteritems(parse_qs(qs)):
            if value and len(value) > 0:
                url_options[name] = value[0]

        # We only support redis:// and unix:// schemes.
        if url.scheme == 'unix':
            url_options.update({
                'password': url.password,
                'path': url.path,
                'connection_class': UnixDomainSocketConnection,
            })

        else:
            url_options.update({
                'host': url.hostname,
                'port': int(url.port or 6380),
                'password': url.password,
            })

            # If there's a path argument, use it as the db argument if a
            # querystring value wasn't specified
            if 'db' not in url_options and url.path:
                try:
                    url_options['db'] = int(url.path.replace('/', ''))
                except (AttributeError, ValueError):
                    pass

            if url.scheme == 'lediss':
                url_options['connection_class'] = SSLConnection

        # last shot at the db value
        url_options['db'] = int(url_options.get('db', db or 0))

        # update the arguments from the URL values
        kwargs.update(url_options)
        return cls(**kwargs)

    def __init__(self, connection_class=Connection, max_connections=None,
                 **connection_kwargs):
        self.pid = os.getpid()
        self.connection_class = connection_class
        self.connection_kwargs = connection_kwargs
        self.max_connections = max_connections or 2 ** 31
        self._created_connections = 0
        self._available_connections = []
        self._in_use_connections = set()

    def _checkpid(self):
        if self.pid != os.getpid():
            self.disconnect()
            self.__init__(self.connection_class, self.max_connections,
                          **self.connection_kwargs)

    def get_connection(self, command_name, *keys, **options):
        "Get a connection from the pool"
        self._checkpid()
        try:
            connection = self._available_connections.pop()
        except IndexError:
            connection = self.make_connection()
        self._in_use_connections.add(connection)
        return connection

    def make_connection(self):
        "Create a new connection"
        if self._created_connections >= self.max_connections:
            raise ConnectionError("Too many connections")
        self._created_connections += 1
        return self.connection_class(**self.connection_kwargs)

    def release(self, connection):
        "Releases the connection back to the pool"
        self._checkpid()
        if connection.pid == self.pid:
            self._in_use_connections.remove(connection)
            self._available_connections.append(connection)

    def disconnect(self):
        "Disconnects all connections in the pool"
        all_conns = chain(self._available_connections,
                          self._in_use_connections)
        for connection in all_conns:
            connection.disconnect()


class BlockingConnectionPool(object):
    """
    Thread-safe blocking connection pool::

        >>> from redis.client import Redis
        >>> client = Redis(connection_pool=BlockingConnectionPool())

    It performs the same function as the default
    ``:py:class: ~redis.connection.ConnectionPool`` implementation, in that,
    it maintains a pool of reusable connections that can be shared by
    multiple redis clients (safely across threads if required).

    The difference is that, in the event that a client tries to get a
    connection from the pool when all of connections are in use, rather than
    raising a ``:py:class: ~redis.exceptions.ConnectionError`` (as the default
    ``:py:class: ~redis.connection.ConnectionPool`` implementation does), it
    makes the client wait ("blocks") for a specified number of seconds until
    a connection becomes available.

    Use ``max_connections`` to increase / decrease the pool size::

        >>> pool = BlockingConnectionPool(max_connections=10)

    Use ``timeout`` to tell it either how many seconds to wait for a connection
    to become available, or to block forever:

        # Block forever.
        >>> pool = BlockingConnectionPool(timeout=None)

        # Raise a ``ConnectionError`` after five seconds if a connection is
        # not available.
        >>> pool = BlockingConnectionPool(timeout=5)
    """
    def __init__(self, max_connections=50, timeout=20, connection_class=None,
                 queue_class=None, **connection_kwargs):
        "Compose and assign values."
        # Compose.
        if connection_class is None:
            connection_class = Connection
        if queue_class is None:
            queue_class = LifoQueue

        # Assign.
        self.connection_class = connection_class
        self.connection_kwargs = connection_kwargs
        self.queue_class = queue_class
        self.max_connections = max_connections
        self.timeout = timeout

        # Validate the ``max_connections``.  With the "fill up the queue"
        # algorithm we use, it must be a positive integer.
        is_valid = isinstance(max_connections, int) and max_connections > 0
        if not is_valid:
            raise ValueError('``max_connections`` must be a positive integer')

        # Get the current process id, so we can disconnect and reinstantiate if
        # it changes.
        self.pid = os.getpid()

        # Create and fill up a thread safe queue with ``None`` values.
        self.pool = self.queue_class(max_connections)
        while True:
            try:
                self.pool.put_nowait(None)
            except Full:
                break

        # Keep a list of actual connection instances so that we can
        # disconnect them later.
        self._connections = []

    def _checkpid(self):
        """
        Check the current process id.  If it has changed, disconnect and
        re-instantiate this connection pool instance.
        """
        # Get the current process id.
        pid = os.getpid()

        # If it hasn't changed since we were instantiated, then we're fine, so
        # just exit, remaining connected.
        if self.pid == pid:
            return

        # If it has changed, then disconnect and re-instantiate.
        self.disconnect()
        self.reinstantiate()

    def make_connection(self):
        "Make a fresh connection."
        connection = self.connection_class(**self.connection_kwargs)
        self._connections.append(connection)
        return connection

    def get_connection(self, command_name, *keys, **options):
        """
        Get a connection, blocking for ``self.timeout`` until a connection
        is available from the pool.

        If the connection returned is ``None`` then creates a new connection.
        Because we use a last-in first-out queue, the existing connections
        (having been returned to the pool after the initial ``None`` values
        were added) will be returned before ``None`` values. This means we only
        create new connections when we need to, i.e.: the actual number of
        connections will only increase in response to demand.
        """
        # Make sure we haven't changed process.
        self._checkpid()

        # Try and get a connection from the pool. If one isn't available within
        # self.timeout then raise a ``ConnectionError``.
        connection = None
        try:
            connection = self.pool.get(block=True, timeout=self.timeout)
        except Empty:
            # Note that this is not caught by the redis client and will be
            # raised unless handled by application code. If you want never to
            raise ConnectionError("No connection available.")

        # If the ``connection`` is actually ``None`` then that's a cue to make
        # a new connection to add to the pool.
        if connection is None:
            connection = self.make_connection()

        return connection

    def release(self, connection):
        "Releases the connection back to the pool."
        # Make sure we haven't changed process.
        self._checkpid()

        # Put the connection back into the pool.
        try:
            self.pool.put_nowait(connection)
        except Full:
            # This shouldn't normally happen but might perhaps happen after a
            # reinstantiation. So, we can handle the exception by not putting
            # the connection back on the pool, because we definitely do not
            # want to reuse it.
            pass

    def disconnect(self):
        "Disconnects all connections in the pool."
        for connection in self._connections:
            connection.disconnect()

    def reinstantiate(self):
        """
        Reinstatiate this instance within a new process with a new connection
        pool set.
        """
        self.__init__(max_connections=self.max_connections,
                      timeout=self.timeout,
                      connection_class=self.connection_class,
                      queue_class=self.queue_class, **self.connection_kwargs)



# coding: utf-8
# Test Cases for hash commands

import unittest
import sys
import datetime, time
sys.path.append('..')

import ledis
from ledis._compat import b, iteritems, itervalues
from ledis import ResponseError


l = ledis.Ledis(port=6380)

def current_time():
    return datetime.datetime.now()


class TestCmdHash(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        l.hmclear('myhash', 'a')
        

    def test_hdel(self):
        l.hset('myhash', 'field1', 'foo')
        assert l.hdel('myhash', 'field1') == 1
        assert l.hdel('myhash', 'field1') == 0
        assert l.hdel('myhash', 'field1', 'field2') == 0

    def test_hexists(self):
        l.hset('myhash', 'field1', 'foo')
        l.hdel('myhash', 'field2')
        assert l.hexists('myhash', 'field1') == 1
        assert l.hexists('myhash', 'field2') == 0      

    def test_hget(self):
        l.hset('myhash', 'field1', 'foo')
        assert l.hget('myhash', 'field1') == 'foo'
        assert (l.hget('myhash', 'field2')) is None

    def test_hgetall(self):
        h = {'field1': 'foo', 'field2': 'bar'}
        l.hmset('myhash', h)
        assert l.hgetall('myhash') == h

    def test_hincrby(self):
        assert l.hincrby('myhash', 'field1') == 1
        l.hclear('myhash')
        assert l.hincrby('myhash', 'field1', 1) == 1
        assert l.hincrby('myhash', 'field1', 5) == 6
        assert l.hincrby('myhash', 'field1', -10) == -4

    def test_hkeys(self):
        h = {'field1': 'foo', 'field2': 'bar'}
        l.hmset('myhash', h)
        assert l.hkeys('myhash') == ['field1', 'field2'] 

    def test_hlen(self):
        l.hset('myhash', 'field1', 'foo')
        assert l.hlen('myhash') == 1
        l.hset('myhash', 'field2', 'bar')
        assert l.hlen('myhash') == 2


    def test_hmget(self):
        assert l.hmset('myhash', {'a': '1', 'b': '2', 'c': '3'})
        assert l.hmget('myhash', 'a', 'b', 'c') == ['1', '2', '3']


    def test_hmset(self):
        h = {'a': '1', 'b': '2', 'c': '3'}
        assert l.hmset('myhash', h)
        assert l.hgetall('myhash') == h

    def test_hset(self):
        l.hclear('myhash')
        assert int(l.hset('myhash', 'field1', 'foo')) == 1
        assert l.hset('myhash', 'field1', 'foo') == 0

    def test_hvals(self):
        h = {'a': '1', 'b': '2', 'c': '3'}
        l.hmset('myhash', h)
        local_vals = list(itervalues(h))
        remote_vals = l.hvals('myhash')
        assert sorted(local_vals) == sorted(remote_vals)


    def test_hclear(self):
        h = {'a': '1', 'b': '2', 'c': '3'}
        l.hmset('myhash', h)
        assert l.hclear('myhash') == 3
        assert l.hclear('myhash') == 0


    def test_hmclear(self):
        h = {'a': '1', 'b': '2', 'c': '3'}
        l.hmset('myhash1', h)
        l.hmset('myhash2', h)
        assert l.hmclear('myhash1', 'myhash2') == 2


    def test_hexpire(self):
        assert l.hexpire('myhash', 100) == 0
        l.hset('myhash', 'field1', 'foo')
        assert l.hexpire('myhash', 100) == 1
        assert l.httl('myhash') <= 100

    def test_hexpireat_datetime(self):
        expire_at = current_time() + datetime.timedelta(minutes=1)
        l.hset('a', 'f', 'foo')
        assert l.hexpireat('a', expire_at)
        assert 0 < l.httl('a') <= 61

    def test_hexpireat_unixtime(self):
        expire_at = current_time() + datetime.timedelta(minutes=1)
        l.hset('a', 'f', 'foo')
        expire_at_seconds = int(time.mktime(expire_at.timetuple()))
        assert l.hexpireat('a', expire_at_seconds)
        assert 0 < l.httl('a') <= 61

    def test_zexpireat_no_key(self):
        expire_at = current_time() + datetime.timedelta(minutes=1)
        assert not l.hexpireat('a', expire_at)

    def test_hexpireat(self):
        assert l.hexpireat('myhash', 1577808000) == 0
        l.hset('myhash', 'field1', 'foo')
        assert l.hexpireat('myhash', 1577808000) == 1

    def test_httl(self):
        l.hset('myhash', 'field1', 'foo')
        assert l.hexpire('myhash', 100)
        assert l.httl('myhash') <= 100

    def test_hpersist(self):
        l.hset('myhash', 'field1', 'foo')
        l.hexpire('myhash', 100)
        assert l.httl('myhash') <= 100
        assert l.hpersist('myhash')
        assert l.httl('myhash') == -1




# coding: utf-8
# Test Cases for other commands

import unittest
import sys
sys.path.append('..')

import ledis
from ledis._compat import b
from ledis import ResponseError

l = ledis.Ledis(port=6380)

class TestOtherCommands(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    # server information
    def test_echo(self):
        assert l.echo('foo bar') == b('foo bar')

    def test_ping(self):
        assert l.ping()

    def test_select(self):
        assert l.select('1')
        assert l.select('15')
        self.assertRaises(ResponseError, lambda: l.select('16'))


# coding: utf-8
# Test Cases for zset commands

import unittest
import sys
import datetime, time
sys.path.append('..')

import ledis
from ledis._compat import b, iteritems
from ledis import ResponseError

l = ledis.Ledis(port=6380)

def current_time():
    return datetime.datetime.now()

class TestCmdZset(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        l.zclear('a')

    def test_zadd(self):
        l.zadd('a', a1=1, a2=2, a3=3)
        assert l.zrange('a', 0, -1) == [b('a1'), b('a2'), b('a3')]

    def test_zcard(self):
        l.zadd('a', a1=1, a2=2, a3=3)
        assert l.zcard('a') == 3

    def test_zcount(self):
        l.zadd('a', a1=1, a2=2, a3=3)
        assert l.zcount('a', '-inf', '+inf') == 3
        assert l.zcount('a', 1, 2) == 2
        assert l.zcount('a', 10, 20) == 0

    def test_zincrby(self):
        l.zadd('a', a1=1, a2=2, a3=3)
        assert l.zincrby('a', 'a2') == 3
        assert l.zincrby('a', 'a3', amount=5) == 8
        assert l.zscore('a', 'a2') == 3
        assert l.zscore('a', 'a3') == 8

    def test_zrange(self):
        l.zadd('a', a1=1, a2=2, a3=3)
        assert l.zrange('a', 0, 1) == [b('a1'), b('a2')]
        assert l.zrange('a', 2, 3) == [b('a3')]

        #withscores
        assert l.zrange('a', 0, 1, withscores=True) == \
            [(b('a1'), 1), (b('a2'), 2)]
        assert l.zrange('a', 2, 3, withscores=True) == \
            [(b('a3'), 3)]

    def test_zrangebyscore(self):
        l.zadd('a', a1=1, a2=2, a3=3, a4=4, a5=5)
        assert l.zrangebyscore('a', 2, 4) == [b('a2'), b('a3'), b('a4')]

        # slicing with start/num
        assert l.zrangebyscore('a', 2, 4, start=1, num=2) == \
            [b('a3'), b('a4')]

        # withscores 
        assert l.zrangebyscore('a', 2, 4, withscores=True) == \
            [('a2', 2), ('a3', 3), ('a4', 4)]

    def test_zrank(self):
        l.zadd('a', a1=1, a2=2, a3=3, a4=4, a5=5)
        assert l.zrank('a', 'a1') == 0
        assert l.zrank('a', 'a3') == 2
        assert l.zrank('a', 'a6') is None

    def test_zrem(self):
        l.zadd('a', a1=1, a2=2, a3=3)
        assert l.zrem('a', 'a2') == 1
        assert l.zrange('a', 0, -1) == [b('a1'), b('a3')]
        assert l.zrem('a', 'b') == 0
        assert l.zrange('a', 0, -1) == [b('a1'), b('a3')]

        # multiple keys
        l.zadd('a', a1=1, a2=2, a3=3)
        assert l.zrem('a', 'a1', 'a2') == 2
        assert l.zrange('a', 0, -1) == [b('a3')]

    def test_zremrangebyrank(self):
        l.zadd('a', a1=1, a2=2, a3=3, a4=4, a5=5)
        assert l.zremrangebyrank('a', 1, 3) == 3
        assert l.zrange('a', 0, -1) == [b('a1'), b('a5')]

    def test_zremrangebyscore(self):
        l.zadd('a', a1=1, a2=2, a3=3, a4=4, a5=5)
        assert l.zremrangebyscore('a', 2, 4) == 3
        assert l.zrange('a', 0, -1) == [b('a1'), b('a5')]
        assert l.zremrangebyscore('a', 2, 4) == 0
        assert l.zrange('a', 0, -1) == [b('a1'), b('a5')]

    def test_zrevrange(self):
        l.zadd('a', a1=1, a2=2, a3=3)
        assert l.zrevrange('a', 0, 1) == [b('a3'), b('a2')]
        assert l.zrevrange('a', 1, 2) == [b('a2'), b('a1')]

    def test_zrevrank(self):
        l.zadd('a', a1=1, a2=2, a3=3, a4=4, a5=5)
        assert l.zrevrank('a', 'a1') == 4
        assert l.zrevrank('a', 'a2') == 3
        assert l.zrevrank('a', 'a6') is None

    def test_zrevrangebyscore(self):
        l.zadd('a', a1=1, a2=2, a3=3, a4=4, a5=5)
        assert l.zrevrangebyscore('a', 4, 2) == [b('a4'), b('a3'), b('a2')]

        # slicing with start/num
        assert l.zrevrangebyscore('a', 4, 2, start=1, num=2) == \
                [b('a3'), b('a2')]

        # withscores
        assert l.zrevrangebyscore('a', 4, 2, withscores=True) == \
                [(b('a4'), 4), (b('a3'), 3), (b('a2'), 2)]

    def test_zscore(self):
        l.zadd('a', a1=1, a2=2, a3=3)
        assert l.zscore('a', 'a1') == 1
        assert l.zscore('a', 'a2') == 2
        assert l.zscore('a', 'a4') is None

    def test_zclear(self):
        l.zadd('a', a1=1, a2=2, a3=3)
        assert l.zclear('a') == 3
        assert l.zclear('a') == 0

    def test_zmclear(self):
        l.zadd('a', a1=1, a2=2, a3=3)
        l.zadd('b', b1=1, b2=2, b3=3)
        assert l.lmclear('a', 'b') == 2
        assert l.lmclear('c', 'd') == 2
 
    def test_zexpire(self):
        assert not l.zexpire('a', 100)
        l.zadd('a', a1=1, a2=2, a3=3)
        assert l.zexpire('a', 100)
        assert 0 < l.zttl('a') <= 100
        assert l.zpersist('a')
        assert l.zttl('a') == -1

    def test_zexpireat_datetime(self):
        expire_at = current_time() + datetime.timedelta(minutes=1)
        l.zadd('a', a1=1)
        assert l.zexpireat('a', expire_at)
        assert 0 < l.zttl('a') <= 61

    def test_zexpireat_unixtime(self):
        expire_at = current_time() + datetime.timedelta(minutes=1)
        l.zadd('a', a1=1)
        expire_at_seconds = int(time.mktime(expire_at.timetuple()))
        assert l.zexpireat('a', expire_at_seconds)
        assert 0 < l.zttl('a') <= 61

    def test_zexpireat_no_key(self):
        expire_at = current_time() + datetime.timedelta(minutes=1)
        assert not l.zexpireat('a', expire_at)

    def test_zttl_and_zpersist(self):
        l.zadd('a', a1=1)
        l.zexpire('a', 100)
        assert 0 < l.zttl('a') <= 100
        assert l.zpersist('a')
        assert l.zttl('a') == -1



"""Internal module for Python 2 backwards compatibility."""
import sys


if sys.version_info[0] < 3:
    from urlparse import parse_qs, urlparse
    from itertools import imap, izip
    from string import letters as ascii_letters
    from Queue import Queue
    try:
        from cStringIO import StringIO as BytesIO
    except ImportError:
        from StringIO import StringIO as BytesIO

    iteritems = lambda x: x.iteritems()
    iterkeys = lambda x: x.iterkeys()
    itervalues = lambda x: x.itervalues()
    nativestr = lambda x: \
        x if isinstance(x, str) else x.encode('utf-8', 'replace')
    u = lambda x: x.decode()
    b = lambda x: x
    next = lambda x: x.next()
    byte_to_chr = lambda x: x
    unichr = unichr
    xrange = xrange
    basestring = basestring
    unicode = unicode
    bytes = str
    long = long
else:
    from urllib.parse import parse_qs, urlparse
    from io import BytesIO
    from string import ascii_letters
    from queue import Queue

    iteritems = lambda x: iter(x.items())
    iterkeys = lambda x: iter(x.keys())
    itervalues = lambda x: iter(x.values())
    byte_to_chr = lambda x: chr(x)
    nativestr = lambda x: \
        x if isinstance(x, str) else x.decode('utf-8', 'replace')
    u = lambda x: x
    b = lambda x: x.encode('iso-8859-1') if not isinstance(x, bytes) else x
    next = next
    unichr = chr
    imap = map
    izip = zip
    xrange = range
    basestring = str
    unicode = str
    bytes = bytes
    long = int

try:  # Python 3
    from queue import LifoQueue, Empty, Full
except ImportError:
    from Queue import Empty, Full
    try:  # Python 2.6 - 2.7
        from Queue import LifoQueue
    except ImportError:  # Python 2.5
        from Queue import Queue
        # From the Python 2.7 lib. Python 2.5 already extracted the core
        # methods to aid implementating different queue organisations.

        class LifoQueue(Queue):
            "Override queue methods to implement a last-in first-out queue."

            def _init(self, maxsize):
                self.maxsize = maxsize
                self.queue = []

            def _qsize(self, len=len):
                return len(self.queue)

            def _put(self, item):
                self.queue.append(item)

            def _get(self):
                return self.queue.pop()



"Core exceptions raised by the LedisDB client"


class LedisError(Exception):
    pass

class ServerError(LedisError):
    pass


class ConnectionError(ServerError):
    pass


class BusyLoadingError(ConnectionError):
    pass


class InvalidResponse(ServerError):
    pass


class ResponseError(LedisError):
    pass


class DataError(LedisError):
    pass


class ExecAbortError(ResponseError):
    pass



# coding: utf-8
# Test Cases for k/v commands

import unittest
import sys
import datetime, time
sys.path.append('..')

import ledis
from ledis._compat import b, iteritems

l = ledis.Ledis(port=6380)


def current_time():
    return datetime.datetime.now()


class TestCmdKv(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        l.delete('a', 'b', 'c', 'non_exist_key')

    def test_decr(self):
        assert l.delete('a') == 1
        assert l.decr('a') == -1
        assert l['a'] == b('-1')
        assert l.decr('a') == -2
        assert l['a'] == b('-2')
        assert l.decr('a', amount=5) == -7
        assert l['a'] == b('-7')

        #FIXME: how to test exception?
        # l.set('b', '234293482390480948029348230948')
        # self.assertRaises(ResponseError, l.delete('b'))

    def test_decrby(self):
        assert l.delete('a') == 1
        assert l.decrby('a') == -1
        assert l['a'] == b('-1')
        assert l.decrby('a') == -2
        assert l['a'] == b('-2')
        assert l.decrby('a', amount=5) == -7
        assert l['a'] == b('-7')

    def test_del(self):
        assert l.delete('a') == 1
        assert l.delete('a', 'b', 'c') == 3

    def test_exists(self):
        l.delete('a', 'non_exist_key')
        l.set('a', 'hello')
        assert (l.exists('a'))
        assert not (l.exists('non_exist_key'))

    def test_get(self):
        l.set('a', 'hello')
        assert l.get('a') == 'hello'
        l.set('b', '中文')
        assert l.get('b') == '中文'
        l.delete('non_exist_key')
        assert (l.get('non_exist_key')) is None

    def test_getset(self):
        l.set('a', 'hello')
        assert l.getset('a', 'world') == 'hello'
        assert l.get('a') == 'world'
        l.delete('non_exist_key')
        assert (l.getset('non_exist_key', 'non')) is None

    def test_incr(self):
        l.delete('non_exist_key')
        assert l.incr('non_exist_key') == 1
        l.set('a', 100)
        assert l.incr('a') == 101

    def test_incrby(self):
        l.delete('a')
        assert l.incrby('a', 100) == 100

        l.set('a', 100)
        assert l.incrby('a', 100) == 200
        assert l.incrby('a', amount=100) == 300

    def test_mget(self):
        l.set('a', 'hello')
        l.set('b', 'world')
        l.delete('non_exist_key')
        assert l.mget('a', 'b', 'non_exist_key') == ['hello', 'world', None]
        l.delete('a', 'b')
        assert l.mget(['a', 'b']) == [None, None]

    def test_mset(self):
        d = {'a': b('1'), 'b': b('2'), 'c': b('3')}
        assert l.mset(**d)
        for k, v in iteritems(d):
            assert l[k] == v

    def test_set(self):
        assert (l.set('a', 100))

    def test_setnx(self):
        l.delete('a')
        assert l.setnx('a', '1')
        assert l['a'] == b('1')
        assert not l.setnx('a', '2')
        assert l['a'] == b('1')

    def test_ttl(self):
        assert l.set('a', 'hello')
        assert l.expire('a', 100)
        assert l.ttl('a') <= 100
        l.delete('a')
        assert l.ttl('a') == -1
        l.set('a', 'hello')
        assert l.ttl('a') == -1

    def test_persist(self):
        assert l.set('a', 'hello')
        assert l.expire('a', 100)
        assert l.ttl('a') <= 100
        assert l.persist('a')
        l.delete('non_exist_key')
        assert not l.persist('non_exist_key')

    def test_expire(self):
        assert not l.expire('a', 100)

        l.set('a', 'hello')
        assert (l.expire('a', 100))
        l.delete('a')
        assert not (l.expire('a', 100))

    def test_expireat_datetime(self):
        expire_at = current_time() + datetime.timedelta(minutes=1)
        l.set('a', '1')
        assert l.expireat('a', expire_at)
        assert 0 < l.ttl('a') <= 61

    def test_expireat_unixtime(self):
        expire_at = current_time() + datetime.timedelta(minutes=1)
        l.set('a', '1')
        expire_at_seconds = int(time.mktime(expire_at.timetuple()))
        assert l.expireat('a', expire_at_seconds)
        assert 0 < l.ttl('a') <= 61

    def test_expireat_no_key(self):
        expire_at = current_time() + datetime.timedelta(minutes=1)
        assert not l.expireat('a', expire_at)

    def test_expireat(self):
        l.set('a', 'hello')
        assert (l.expireat('a', 1577808000)) # time is 2020.1.1
        l.delete('a')
        assert not(l.expireat('a', 1577808000))
        


# coding: utf-8
# Test Cases for list commands

import unittest
import datetime, time
import sys
sys.path.append('..')

import ledis
from ledis._compat import b

l = ledis.Ledis(port=6380)


def current_time():
    return datetime.datetime.now()


class TestCmdList(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        l.lmclear('mylist', 'mylist1', 'mylist2')

    def test_lindex(self):
        l.rpush('mylist', '1', '2', '3')
        assert l.lindex('mylist', 0) == b('1')
        assert l.lindex('mylist', 1) == b('2')
        assert l.lindex('mylist', 2) == b('3')

    def test_llen(self):
        l.rpush('mylist', '1', '2', '3')
        assert l.llen('mylist') == 3

    def test_lpop(self):
        l.rpush('mylist', '1', '2', '3')
        assert l.lpop('mylist') == b('1')
        assert l.lpop('mylist') == b('2')
        assert l.lpop('mylist') == b('3')
        assert l.lpop('mylist') is None

    def test_lpush(self):
        assert l.lpush('mylist', '1') == 1
        assert l.lpush('mylist', '2') == 2
        assert l.lpush('mylist', '3', '4', '5') == 5
        assert l.lrange('mylist', 0, -1) == ['5', '4', '3', '2', '1']

    def test_lrange(self):
        l.rpush('mylist', '1', '2', '3', '4', '5')
        assert l.lrange('mylist', 0, 2) == ['1', '2', '3']
        assert l.lrange('mylist', 2, 10) == ['3', '4', '5']
        assert l.lrange('mylist', 0, -1) == ['1', '2', '3', '4', '5']

    def test_rpush(self):
        assert l.rpush('mylist', '1') == 1
        assert l.rpush('mylist', '2') == 2
        assert l.rpush('mylist', '3', '4') == 4
        assert l.lrange('mylist', 0, -1) == ['1', '2', '3', '4']

    def test_rpop(self):
        l.rpush('mylist', '1', '2', '3')
        assert l.rpop('mylist') == b('3')
        assert l.rpop('mylist') == b('2')
        assert l.rpop('mylist') == b('1')
        assert l.rpop('mylist') is None

    def test_lclear(self):
        l.rpush('mylist', '1', '2', '3')
        assert l.lclear('mylist') == 3
        assert l.lclear('mylist') == 0

    def test_lmclear(self):
        l.rpush('mylist1', '1', '2', '3')
        l.rpush('mylist2', '1', '2', '3')
        assert l.lmclear('mylist1', 'mylist2') == 2

    def test_lexpire(self):
        assert not l.lexpire('mylist', 100)
        l.rpush('mylist', '1')
        assert l.lexpire('mylist', 100)
        assert 0 < l.lttl('mylist') <= 100
        assert l.lpersist('mylist')
        assert l.lttl('mylist') == -1

    def test_lexpireat_datetime(self):
        expire_at = current_time() + datetime.timedelta(minutes=1)
        l.rpush('mylist', '1')
        assert l.lexpireat('mylist', expire_at)
        assert 0 < l.lttl('mylist') <= 61

    def test_lexpireat_unixtime(self):
        expire_at = current_time() + datetime.timedelta(minutes=1)
        l.rpush('mylist', '1')
        expire_at_seconds = int(time.mktime(expire_at.timetuple()))
        assert l.lexpireat('mylist', expire_at_seconds)
        assert l.lttl('mylist') <= 61

    def test_lexpireat_no_key(self):
        expire_at = current_time() + datetime.timedelta(minutes=1)
        assert not l.lexpireat('mylist', expire_at)

    def test_lttl_and_lpersist(self):
        l.rpush('mylist', '1')
        l.lexpire('mylist', 100)
        assert 0 < l.lttl('mylist') <= 100
        assert l.lpersist('mylist')
        assert l.lttl('mylist') == -1




from ledis.client import Ledis
from ledis.connection import (
    BlockingConnectionPool,
    ConnectionPool,
    Connection,
    UnixDomainSocketConnection
)
from ledis.utils import from_url
from ledis.exceptions import (
    ConnectionError,
    BusyLoadingError,
    DataError,
    InvalidResponse,
    LedisError,
    ResponseError,
)


__version__ = '0.0.1'
VERSION = tuple(map(int, __version__.split('.')))

__all__ = [
    'Ledis', 'ConnectionPool', 'BlockingConnectionPool',
    'Connection', 'UnixDomainSocketConnection',
    'LedisError', 'ConnectionError', 'ResponseError', 
    'InvalidResponse', 'DataError', 'from_url',  'BusyLoadingError',
]



#!/usr/bin/env python
import os
import sys

from ledis import __version__

try:
    from setuptools import setup
    from setuptools.command.test import test as TestCommand

    class PyTest(TestCommand):
        def finalize_options(self):
            TestCommand.finalize_options(self)
            self.test_args = []
            self.test_suite = True

        def run_tests(self):
            # import here, because outside the eggs aren't loaded
            import pytest
            errno = pytest.main(self.test_args)
            sys.exit(errno)

except ImportError:

    from distutils.core import setup
    PyTest = lambda x: x


setup(
    name='ledis',
    version=__version__,
    description='Python client for ledisdb key-value database',
    long_description='Python client for ledisdb key-value database',
    url='https://github.com/siddontang/ledisdb',
    keywords=['ledis', 'key-value store'],
    license='MIT',
    packages=['ledis'],
    tests_require=['pytest>=2.5.0'],
    cmdclass={'test': PyTest},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ]
)



from __future__ import with_statement
import datetime
import time as mod_time
from ledis._compat import (b, izip, imap, iteritems, iterkeys, itervalues,
                           basestring, long, nativestr, urlparse, bytes)
from ledis.connection import ConnectionPool, UnixDomainSocketConnection
from ledis.exceptions import (
    ConnectionError,
    DataError,
    LedisError,
    ResponseError,
    ExecAbortError,
)

SYM_EMPTY = b('')


def list_or_args(keys, args):
    # returns a single list combining keys and args
    try:
        iter(keys)
        # a string or bytes instance can be iterated, but indicates
        # keys wasn't passed as a list
        if isinstance(keys, (basestring, bytes)):
            keys = [keys]
    except TypeError:
        keys = [keys]
    if args:
        keys.extend(args)
    return keys


def string_keys_to_dict(key_string, callback):
    return dict.fromkeys(key_string.split(), callback)


def dict_merge(*dicts):
    merged = {}
    [merged.update(d) for d in dicts]
    return merged


def pairs_to_dict(response):
    "Create a dict given a list of key/value pairs"
    it = iter(response)
    return dict(izip(it, it))


def zset_score_pairs(response, **options):
    """
    If ``withscores`` is specified in the options, return the response as
    a list of (value, score) pairs
    """
    if not response or not options['withscores']:
        return response
    score_cast_func = options.get('score_cast_func', int)
    it = iter(response)
    return list(izip(it, imap(score_cast_func, it)))


def int_or_none(response):
    if response is None:
        return None
    return int(response)


class Ledis(object):
    """
    Implementation of the Redis protocol.

    This abstract class provides a Python interface to all LedisDB commands
    and an implementation of the Redis protocol.

    Connection and Pipeline derive from this, implementing how
    the commands are sent and received to the Ledis server
    """
    RESPONSE_CALLBACKS = dict_merge(
        string_keys_to_dict(
            'EXISTS EXPIRE EXPIREAT HEXISTS HMSET SETNX '
            'PERSIST HPERSIST LPERSIST ZPERSIST',
            bool
        ),
        string_keys_to_dict(
            'DECRBY DEL HDEL HLEN INCRBY LLEN '
            'ZADD ZCARD ZREM ZREMRANGEBYRANK ZREMRANGEBYSCORE'
            'LMCLEAR HMCLEAR ZMCLEAR',
            int
        ),
        string_keys_to_dict(
            'LPUSH RPUSH',
            lambda r: isinstance(r, long) and r or nativestr(r) == 'OK'
        ),
        string_keys_to_dict(
            'MSET SELECT ',
            lambda r: nativestr(r) == 'OK'
        ),
        string_keys_to_dict(
            'ZRANGE ZRANGEBYSCORE ZREVRANGE ZREVRANGEBYSCORE',
            zset_score_pairs
        ),
        string_keys_to_dict('ZRANK ZREVRANK ZSCORE ZINCRBY', int_or_none),
        {
            'HGETALL': lambda r: r and pairs_to_dict(r) or {},
            'PING': lambda r: nativestr(r) == 'PONG',
            'SET': lambda r: r and nativestr(r) == 'OK', }
    )

    @classmethod
    def from_url(cls, url, db=None, **kwargs):
        """
        Return a Ledis client object configured from the given URL.

        For example::

            redis://[:password]@localhost:6380/0
            unix://[:password]@/path/to/socket.sock?db=0

        There are several ways to specify a database number. The parse function
        will return the first specified option:
            1. A ``db`` querystring option, e.g. redis://localhost?db=0
            2. If using the redis:// scheme, the path argument of the url, e.g.
               redis://localhost/0
            3. The ``db`` argument to this function.

        If none of these options are specified, db=0 is used.

        Any additional querystring arguments and keyword arguments will be
        passed along to the ConnectionPool class's initializer. In the case
        of conflicting arguments, querystring arguments always win.
        """
        connection_pool = ConnectionPool.from_url(url, db=db, **kwargs)
        return cls(connection_pool=connection_pool)

    def __init__(self, host='localhost', port=6380,
                 db=0, socket_timeout=None,
                 connection_pool=None, charset='utf-8',
                 errors='strict', decode_responses=False,
                 unix_socket_path=None):
        if not connection_pool:
            kwargs = {
                'db': db,
                'socket_timeout': socket_timeout,
                'encoding': charset,
                'encoding_errors': errors,
                'decode_responses': decode_responses,
            }
            # based on input, setup appropriate connection args
            if unix_socket_path:
                kwargs.update({
                    'path': unix_socket_path,
                    'connection_class': UnixDomainSocketConnection
                })
            else:
                kwargs.update({
                    'host': host,
                    'port': port
                })
            connection_pool = ConnectionPool(**kwargs)
        self.connection_pool = connection_pool
        self.response_callbacks = self.__class__.RESPONSE_CALLBACKS.copy()

    def set_response_callback(self, command, callback):
        "Set a custom Response Callback"
        self.response_callbacks[command] = callback

    #### COMMAND EXECUTION AND PROTOCOL PARSING ####

    def execute_command(self, *args, **options):
        "Execute a command and return a parsed response"
        pool = self.connection_pool
        command_name = args[0]
        connection = pool.get_connection(command_name, **options)
        try:
            connection.send_command(*args)
            return self.parse_response(connection, command_name, **options)
        except ConnectionError:
            connection.disconnect()
            connection.send_command(*args)
            return self.parse_response(connection, command_name, **options)
        finally:
            pool.release(connection)

    def parse_response(self, connection, command_name, **options):
        "Parses a response from the Ledis server"
        response = connection.read_response()
        if command_name in self.response_callbacks:
            return self.response_callbacks[command_name](response, **options)
        return response


    #### SERVER INFORMATION ####
    def echo(self, value):
        "Echo the string back from the server"
        return self.execute_command('ECHO', value)

    def ping(self):
        "Ping the Ledis server"
        return self.execute_command('PING')

    def select(self, db):
        """Select a Ledis db, ``db`` is integer type"""
        try:
            db = int(db)
        except ValueError:
            db = 0
        return self.execute_command('SELECT', db)


    #### BASIC KEY COMMANDS ####
    def decr(self, name, amount=1):
        """
        Decrements the value of ``key`` by ``amount``.  If no key exists,
        the value will be initialized as 0 - ``amount``
        """
        return self.execute_command('DECRBY', name, amount)

    def decrby(self, name, amount=1):
        """
        Decrements the value of ``key`` by ``amount``.  If no key exists,
        the value will be initialized as 0 - ``amount``
        """
        return self.decr(name, amount)

    def delete(self, *names):
        "Delete one or more keys specified by ``names``"
        return self.execute_command('DEL', *names)

    def exists(self, name):
        "Returns a boolean indicating whether key ``name`` exists"
        return self.execute_command('EXISTS', name)

    def expire(self, name, time):
        """
        Set an expire flag on key ``name`` for ``time`` seconds. ``time``
        can be represented by an integer or a Python timedelta object.
        """
        if isinstance(time, datetime.timedelta):
            time = time.seconds + time.days * 24 * 3600
        return self.execute_command('EXPIRE', name, time)

    def expireat(self, name, when):
        """
        Set an expire flag on key ``name``. ``when`` can be represented
        as an integer indicating unix time or a Python datetime object.
        """
        if isinstance(when, datetime.datetime):
            when = int(mod_time.mktime(when.timetuple()))
        return self.execute_command('EXPIREAT', name, when)

    def get(self, name):
        """
        Return the value at key ``name``, or None if the key doesn't exist
        """
        return self.execute_command('GET', name)

    def __getitem__(self, name):
        """
        Return the value at key ``name``, raises a KeyError if the key
        doesn't exist.
        """
        value = self.get(name)
        if value:
            return value
        raise KeyError(name)

    def getset(self, name, value):
        """
        Set the value at key ``name`` to ``value`` if key doesn't exist
        Return the value at key ``name`` atomically
        """
        return self.execute_command('GETSET', name, value)

    def incr(self, name, amount=1):
        """
        Increments the value of ``key`` by ``amount``.  If no key exists,
        the value will be initialized as ``amount``
        """
        return self.execute_command('INCRBY', name, amount)

    def incrby(self, name, amount=1):
        """
        Increments the value of ``key`` by ``amount``.  If no key exists,
        the value will be initialized as ``amount``
        """

        # An alias for ``incr()``, because it is already implemented
        # as INCRBY ledis command.
        return self.incr(name, amount)

    def mget(self, keys, *args):
        """
        Returns a list of values ordered identically to ``keys``
        """
        args = list_or_args(keys, args)
        return self.execute_command('MGET', *args)

    def mset(self, *args, **kwargs):
        """
        Sets key/values based on a mapping. Mapping can be supplied as a single
        dictionary argument or as kwargs.
        """
        if args:
            if len(args) != 1 or not isinstance(args[0], dict):
                raise LedisError('MSET requires **kwargs or a single dict arg')
            kwargs.update(args[0])
        items = []
        for pair in iteritems(kwargs):
            items.extend(pair)
        return self.execute_command('MSET', *items)

    def set(self, name, value):
        """
        Set the value of key ``name`` to ``value``.
        """
        pieces = [name, value]
        return self.execute_command('SET', *pieces)

    def setnx(self, name, value):
        "Set the value of key ``name`` to ``value`` if key doesn't exist"
        return self.execute_command('SETNX', name, value)

    def ttl(self, name):
        "Returns the number of seconds until the key ``name`` will expire"
        return self.execute_command('TTL', name)

    def persist(self, name):
        "Removes an expiration on name"
        return self.execute_command('PERSIST', name)


    #### LIST COMMANDS ####
    def lindex(self, name, index):
        """
        Return the item from list ``name`` at position ``index``

        Negative indexes are supported and will return an item at the
        end of the list
        """
        return self.execute_command('LINDEX', name, index)

    def llen(self, name):
        "Return the length of the list ``name``"
        return self.execute_command('LLEN', name)

    def lpop(self, name):
        "Remove and return the first item of the list ``name``"
        return self.execute_command('LPOP', name)

    def lpush(self, name, *values):
        "Push ``values`` onto the head of the list ``name``"
        return self.execute_command('LPUSH', name, *values)

    def lrange(self, name, start, end):
        """
        Return a slice of the list ``name`` between
        position ``start`` and ``end``

        ``start`` and ``end`` can be negative numbers just like
        Python slicing notation
        """
        return self.execute_command('LRANGE', name, start, end)

    def rpop(self, name):
        "Remove and return the last item of the list ``name``"
        return self.execute_command('RPOP', name)

    def rpush(self, name, *values):
        "Push ``values`` onto the tail of the list ``name``"
        return self.execute_command('RPUSH', name, *values)

    # SPECIAL COMMANDS SUPPORTED BY LEDISDB
    def lclear(self, name):
        "Delete the key of ``name``"
        return self.execute_command("LCLEAR", name)

    def lmclear(self, *names):
        "Delete multiple keys of ``name``"
        return self.execute_command('LMCLEAR', *names)

    def lexpire(self, name, time):
        """
        Set an expire flag on key ``name`` for ``time`` seconds. ``time``
        can be represented by an integer or a Python timedelta object.
        """
        if isinstance(time, datetime.timedelta):
            time = time.seconds + time.days * 24 * 3600
        return self.execute_command("LEXPIRE", name, time)

    def lexpireat(self, name, when):
        """
        Set an expire flag on key ``name``. ``when`` can be represented as an integer
        indicating unix time or a Python datetime object.
        """
        if isinstance(when, datetime.datetime):
            when = int(mod_time.mktime(when.timetuple()))
        return self.execute_command('LEXPIREAT', name, when)

    def lttl(self, name):
        "Returns the number of seconds until the key ``name`` will expire"
        return self.execute_command('LTTL', name)

    def lpersist(self, name):
        "Removes an expiration on ``name``"
        return self.execute_command('LPERSIST', name)


    #### SORTED SET COMMANDS ####
    def zadd(self, name, *args, **kwargs):
        """
        Set any number of score, element-name pairs to the key ``name``. Pairs
        can be specified in two ways:

        As *args, in the form of: score1, name1, score2, name2, ...
        or as **kwargs, in the form of: name1=score1, name2=score2, ...

        The following example would add four values to the 'my-key' key:
        ledis.zadd('my-key', 1.1, 'name1', 2.2, 'name2', name3=3.3, name4=4.4)
        """
        pieces = []
        if args:
            if len(args) % 2 != 0:
                raise LedisError("ZADD requires an equal number of "
                                 "values and scores")
            pieces.extend(args)
        for pair in iteritems(kwargs):
            pieces.append(pair[1])
            pieces.append(pair[0])
        return self.execute_command('ZADD', name, *pieces)

    def zcard(self, name):
        "Return the number of elements in the sorted set ``name``"
        return self.execute_command('ZCARD', name)

    def zcount(self, name, min, max):
        """
        Return the number of elements in the sorted set at key ``name`` with a score 
        between ``min`` and ``max``. 
        The min and max arguments have the same semantic as described for ZRANGEBYSCORE.
        """
        return self.execute_command('ZCOUNT', name, min, max)

    def zincrby(self, name, value, amount=1):
        "Increment the score of ``value`` in sorted set ``name`` by ``amount``"
        return self.execute_command('ZINCRBY', name, amount, value)

    def zrange(self, name, start, end, desc=False, withscores=False):
        """
        Return a range of values from sorted set ``name`` between
        ``start`` and ``end`` sorted in ascending order.

        ``start`` and ``end`` can be negative, indicating the end of the range.

        ``desc`` a boolean indicating whether to sort the results descendingly

        ``withscores`` indicates to return the scores along with the values.
        The return type is a list of (value, score) pairs

        ``score_cast_func`` a callable used to cast the score return value
        """
        if desc:
            return self.zrevrange(name, start, end, withscores,
                                  score_cast_func)
        pieces = ['ZRANGE', name, start, end]
        if withscores:
            pieces.append('withscores')
        options = {
            'withscores': withscores, 'score_cast_func': int}
        return self.execute_command(*pieces, **options)

    def zrangebyscore(self, name, min, max, start=None, num=None,
                      withscores=False):
        """
        Return a range of values from the sorted set ``name`` with scores
        between ``min`` and ``max``.

        If ``start`` and ``num`` are specified, then return a slice
        of the range.

        ``withscores`` indicates to return the scores along with the values.
        The return type is a list of (value, score) pairs

        `score_cast_func`` a callable used to cast the score return value
        """
        if (start is not None and num is None) or \
                (num is not None and start is None):
            raise LedisError("``start`` and ``num`` must both be specified")
        pieces = ['ZRANGEBYSCORE', name, min, max]
        if start is not None and num is not None:
            pieces.extend(['LIMIT', start, num])
        if withscores:
            pieces.append('withscores')
        options = {
            'withscores': withscores, 'score_cast_func': int}
        return self.execute_command(*pieces, **options)

    def zrank(self, name, value):
        """
        Returns a 0-based value indicating the rank of ``value`` in sorted set
        ``name``
        """
        return self.execute_command('ZRANK', name, value)

    def zrem(self, name, *values):
        "Remove member ``values`` from sorted set ``name``"
        return self.execute_command('ZREM', name, *values)

    def zremrangebyrank(self, name, min, max):
        """
        Remove all elements in the sorted set ``name`` with ranks between
        ``min`` and ``max``. Values are 0-based, ordered from smallest score
        to largest. Values can be negative indicating the highest scores.
        Returns the number of elements removed
        """
        return self.execute_command('ZREMRANGEBYRANK', name, min, max)

    def zremrangebyscore(self, name, min, max):
        """
        Remove all elements in the sorted set ``name`` with scores
        between ``min`` and ``max``. Returns the number of elements removed.
        """
        return self.execute_command('ZREMRANGEBYSCORE', name, min, max)

    def zrevrange(self, name, start, num, withscores=False):
        """
        Return a range of values from sorted set ``name`` between
        ``start`` and ``num`` sorted in descending order.

        ``start`` and ``num`` can be negative, indicating the end of the range.

        ``withscores`` indicates to return the scores along with the values
        The return type is a list of (value, score) pairs

        ``score_cast_func`` a callable used to cast the score return value
        """
        pieces = ['ZREVRANGE', name, start, num]
        if withscores:
            pieces.append('withscores')
        options = {
            'withscores': withscores, 'score_cast_func': int}
        return self.execute_command(*pieces, **options)

    def zrevrangebyscore(self, name, min, max, start=None, num=None,
                         withscores=False):
        """
        Return a range of values from the sorted set ``name`` with scores
        between ``min`` and ``max`` in descending order.

        If ``start`` and ``num`` are specified, then return a slice
        of the range.

        ``withscores`` indicates to return the scores along with the values.
        The return type is a list of (value, score) pairs

        ``score_cast_func`` a callable used to cast the score return value
        """
        if (start is not None and num is None) or \
                (num is not None and start is None):
            raise LedisError("``start`` and ``num`` must both be specified")
        pieces = ['ZREVRANGEBYSCORE', name, min, max]
        if start is not None and num is not None:
            pieces.extend(['LIMIT', start, num])
        if withscores:
            pieces.append('withscores')
        options = {
            'withscores': withscores, 'score_cast_func': int}
        return self.execute_command(*pieces, **options)

    def zrevrank(self, name, value):
        """
        Returns a 0-based value indicating the descending rank of
        ``value`` in sorted set ``name``
        """
        return self.execute_command('ZREVRANK', name, value)

    def zscore(self, name, value):
        "Return the score of element ``value`` in sorted set ``name``"
        return self.execute_command('ZSCORE', name, value)

    # SPECIAL COMMANDS SUPPORTED BY LEDISDB
    def zclear(self, name):
        "Delete key of ``name`` from sorted set"
        return self.execute_command('ZCLEAR', name)

    def zmclear(self, *names):
        "Delete multiple keys of ``names`` from sorted set"
        return self.execute_command('ZMCLEAR', *names)

    def zexpire(self, name, time):
        "Set timeout on key ``name`` with ``time``"
        if isinstance(time, datetime.timedelta):
            time = time.seconds + time.days * 24 * 3600
        return self.execute_command('ZEXPIRE', name, time)

    def zexpireat(self, name, when):
        """
        Set an expire flag on key name for time seconds. time can be represented by
         an integer or a Python timedelta object.
        """

        if isinstance(when, datetime.datetime):
            when = int(mod_time.mktime(when.timetuple()))
        return self.execute_command('ZEXPIREAT', name, when)

    def zttl(self, name):
        "Returns the number of seconds until the key name will expire"
        return self.execute_command('ZTTL', name)

    def zpersist(self, name):
        "Removes an expiration on name"
        return self.execute_command('ZPERSIST', name)



    #### HASH COMMANDS ####
    def hdel(self, name, *keys):
        "Delete ``keys`` from hash ``name``"
        return self.execute_command('HDEL', name, *keys)

    def hexists(self, name, key):
        "Returns a boolean indicating if ``key`` exists within hash ``name``"
        return self.execute_command('HEXISTS', name, key)

    def hget(self, name, key):
        "Return the value of ``key`` within the hash ``name``"
        return self.execute_command('HGET', name, key)

    def hgetall(self, name):
        "Return a Python dict of the hash's name/value pairs"
        return self.execute_command('HGETALL', name)

    def hincrby(self, name, key, amount=1):
        "Increment the value of ``key`` in hash ``name`` by ``amount``"
        return self.execute_command('HINCRBY', name, key, amount)

    def hkeys(self, name):
        "Return the list of keys within hash ``name``"
        return self.execute_command('HKEYS', name)

    def hlen(self, name):
        "Return the number of elements in hash ``name``"
        return self.execute_command('HLEN', name)

    def hmget(self, name, keys, *args):
        "Returns a list of values ordered identically to ``keys``"
        args = list_or_args(keys, args)
        return self.execute_command('HMGET', name, *args)

    def hmset(self, name, mapping):
        """
        Sets each key in the ``mapping`` dict to its corresponding value
        in the hash ``name``
        """
        if not mapping:
            raise DataError("'hmset' with 'mapping' of length 0")
        items = []
        for pair in iteritems(mapping):
            items.extend(pair)
        return self.execute_command('HMSET', name, *items)

    def hset(self, name, key, value):
        """
        Set ``key`` to ``value`` within hash ``name``
        Returns 1 if HSET created a new field, otherwise 0
        """
        return self.execute_command('HSET', name, key, value)

    def hvals(self, name):
        "Return the list of values within hash ``name``"
        return self.execute_command('HVALS', name)

    # SPECIAL COMMANDS SUPPORTED BY LEDISDB
    def hclear(self, name):
        "Delete key ``name`` from hash"
        return self.execute_command('HCLEAR', name)

    def hmclear(self, *names):
        "Delete multiple keys ``names`` from hash"
        return self.execute_command('HMCLEAR', *names)

    def hexpire(self, name, time):
        """
        Set an expire flag on key name for time milliseconds. 
        time can be represented by an integer or a Python timedelta object.
        """
        if isinstance(time, datetime.timedelta):
            time = time.seconds + time.days * 24 * 3600
        return self.execute_command('HEXPIRE', name, time)

    def hexpireat(self, name, when):
        """
        Set an expire flag on key name. when can be represented as an integer representing 
        unix time in milliseconds (unix time * 1000) or a Python datetime object.
        """
        if isinstance(when, datetime.datetime):
            when = int(mod_time.mktime(when.timetuple()))
        return self.execute_command('HEXPIREAT', name, when)

    def httl(self, name):
        "Returns the number of seconds until the key name will expire"
        return self.execute_command('HTTL', name)

    def hpersist(self, name):
        "Removes an expiration on name"
        return self.execute_command('HPERSIST', name)





def from_url(url, db=None, **kwargs):
    """
    Returns an active Ledis client generated from the given database URL.

    Will attempt to extract the database id from the path url fragment, if
    none is provided.
    """
    from ledis.client import Ledis
    return Ledis.from_url(url, db, **kwargs)
